from database.helper.connection import PerRequestConnection, \
   PerRequestConnectionCursor, log, service_extensions
from bn import AttributeDict
from database.helper import plugin

class DatabaseConnectionService(object):
    def __init__(self, config, param_style='format'):
        self.fetch_converter = None
        if config.has_key('fetch_converter'):
            self.fetch_converter = config['fetch_converter']
            del config['fetch_converter']
        self.execute_converter = None
        if config.has_key('execute_converter'):
            self.execute_converter = config['execute_converter']
            del config['execute_converter']
        connection_args = {}
        for k, v in config.items():
            if k in [
                'database', 
                'user', 
                'host', 
                'password', 
                'port', 
                'fetch_converter',
                'execute_converter',

                'unicode_results', 
                'dsn',
            ]:
                connection_args[str(k)] = v
        if config.get('pool', False) is True:
            from DBUtils.PooledDB import PooledDB
            for k, v in config.items():
                if k in [
                    'creator',
                    'mincached',
                    'maxcached',
                    'maxshared',
                    'maxconnections',
                    'blocking',
                    'maxusage',
                    'setsession',
                ]:
                    connection_args[str(k)] = v
            pool = PooledDB(**connection_args)
            self.connect = pool.connection
        else:
            creator = config.creator
            def connect():
                return creator(**connection_args)
            self.connect = connect
        self.param_style = param_style

    @staticmethod
    def create(flow, name, config=None):
        if config is None:
            config = DatabaseConnectionService.config(flow, name)
        return DatabaseConnectionService(config)

    @staticmethod
    def config(flow, name):
        from configconvert import handle_option_error, handle_section_error
        if not flow.config.option_group.has_key(name):
            raise handle_section_error(
                flow, 
                name, 
                (
                    "'%s.database' or '%s.plugin'"%(name, name)
                )
            )
        from stringconvert import unicodeToUnicode, unicodeToInteger,\
           unicodeToBoolean
        from recordconvert import toRecord
        from configconvert import stringToObject
        from conversionkit import Conversion, chainConverters

        # Re-use the converters   
        unicode_to_integer = unicodeToInteger()
        null = unicodeToUnicode()
    
        connection_service_converter = toRecord(
            missing_defaults=dict(
                pool=False,
                mincached=0,
                maxcached=0,
                maxshared=0,
                maxconnections=0,
                blocking=0,
                maxusage=0,
            ),
            missing_or_empty_errors = dict(
                plugin="The required option '%s.plugin' is missing"%(name,),
            ),
            converters=dict(
                pool = unicodeToBoolean(),
                plugin = null,#stringToObject(),
                mincached = unicode_to_integer,
                maxcached = unicode_to_integer,
                maxshared = unicode_to_integer,
                maxconnections = unicode_to_integer,
                blocking = unicode_to_integer,
                maxusage = unicode_to_integer,
            )
        )
        config = flow.config.option_group[name]
        conversion = Conversion(config).perform(connection_service_converter)
        if not conversion.successful:
            handle_option_error(conversion, name)
        else:
            connection_service_config = conversion.result

        plugin_name = connection_service_config.plugin
        if not plugin.plugins_loaded:
            plugin.load_available_plugins()
        if not plugin.available_plugins.has_key(plugin_name):
            raise Exception('Plugin %r not found'%plugin_name)
        if not plugin.update_config.has_key(plugin_name):
            plugin.update_config[plugin_name] = \
               plugin.available_plugins[plugin_name]['update_config'].load()
        if not plugin.engine_name.has_key(plugin_name):
            plugin.engine_name[plugin_name] = \
               plugin.available_plugins[plugin_name]['engine_name'].load()
        if not plugin.driver_name.has_key(plugin_name):
            plugin.driver_name[plugin_name] = \
               plugin.available_plugins[plugin_name]['driver_name'].load()
        if not plugin.param_style.has_key(plugin_name):
            plugin.param_style[plugin_name] = \
               plugin.available_plugins[plugin_name]['param_style'].load()

        flow.config[name] = plugin.update_config[plugin_name](
            flow,
            name,
            connection_service_config,
        )
        flow.config[name]['plugin'] = plugin_name
        flow.config[name]['engine'] = plugin.engine_name[plugin_name]
        flow.config[name]['driver'] = plugin.driver_name[plugin_name]
        flow.config[name]['param_style'] = plugin.param_style[plugin_name]
        return flow.config[name]

    def start(self, flow, name):
        flow[name] = AttributeDict()
        def connect():
            if not flow[name].has_key('_connection'):
                flow[name]['_connection'] = PerRequestConnection(
                    self.connect(),
                    self.fetch_converter,
                    self.execute_converter,
                )
            return flow[name]['_connection']
        flow[name]['connect'] = connect
        flow[name]['param_style'] = self.param_style
        service_extensions(flow, name)

    def stop(self, flow, name):
        if flow[name].has_key('_connection'):
            if flow[name]._connection._should_commit:
                log.debug('Automatically commiting changes to the database')
                flow[name]._connection._connection.commit()
            flow[name]._connection._connection.close()

    def error(self, flow, name):
        if flow[name].has_key('_connection'):
            log.warning('An error occurred, rolling back database changes')
            flow[name]._connection._should_commit = False
            flow[name]._connection._connection.rollback()


